import builtins
import sys

from radicli import Arg
from wasabi import msg

from .. import ty
from ..cli import cli
from ..errors import BrokerError, CLIError, HTTPXErrors, ProdigyTeamsError
from ..messages import Messages
from ..prodigy_teams_broker_sdk.models import JobLogRequest
from ..prodigy_teams_pam_sdk.models import TaskDetail, TaskSummary
from ..query import delete_job, resolve_recipe, resolve_task, start_job, stop_job
from ..ui import (
    print_args_table,
    print_info_table,
    print_logs,
    print_recipes_help,
    print_table_with_select,
)
from ._recipe_subcommand import create_from_recipe, request_recipes
from ._state import get_auth_state


@cli.subcommand_with_extra(
    "tasks",
    "create",
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok),
    no_start=Arg("--no-start", help=Messages.no_start),
    _show_help=Arg("--help", "-h", help=Messages.help),
)
def create(
    exists_ok: bool = False,
    no_start: bool = False,
    _show_help: bool = False,
    _extra: ty.List[str] = [],
) -> ty.UUID:
    """
    Create a new action. The available action recipes are fetched from your
    cluster and are added as dynamic subcommands. You can see more details
    and available arguments by calling the subcommand with --help, e.g. create
    [name] --help
    """
    # TODO: We should just get one recipe here, rather than all of them. That
    # will also save us from having to query for all the objects.
    auth = get_auth_state()
    try:
        schemas = request_recipes(auth=auth, is_action=False)
    except ProdigyTeamsError as e:
        raise CLIError(Messages.E009, e)
    except HTTPXErrors as e:
        raise CLIError(Messages.E009, e)
    if not _extra:
        print_recipes_help(schemas, "Create a new task", "tasks create")
        sys.exit(0)
    args = [*_extra]
    name = args.pop(0)
    if name not in schemas:
        opts = f"Available: {', '.join(schemas.keys())}"
        raise CLIError(Messages.E010.format(noun="task", name=name), opts)
    schema = resolve_recipe(name, broker_id=auth.broker_id)
    task_id, plan = create_from_recipe(
        schema, args, command="tasks", show_help=_show_help, exists_ok=exists_ok
    )
    msg.good(Messages.T002.format(noun="task", name=task_id))
    print_args_table(plan.args, schema.form_schema.cli_names)
    if not no_start:
        start(task_id)
    return task_id


@cli.subcommand(
    "tasks",
    "list",
    # fmt: off
    select=Arg("--select", help=Messages.select.format(opts=list(TaskSummary.__fields__))),
    as_json=Arg("--json", help=Messages.as_json),
    # fmt: on
)
def list(
    select: ty.List[str] = ["id", "name", "state", "project_name"],
    as_json: bool = False,
) -> ty.Sequence[TaskSummary]:
    """
    List the tasks on the cluster. By default, this includes their ID, name
    and current state, e.g. created or completed
    """
    client = get_auth_state().client
    res = client.task.all(page=1, size=100)
    print_table_with_select(res.items, select=select, as_json=as_json)
    return res.items


@cli.subcommand(
    "tasks",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="task")),
    project_id=Arg(help=Messages.project_id.format(noun="task")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="task")),
    select=Arg(
        "--select",
        help=Messages.select.format(opts=builtins.list(TaskDetail.__fields__)),
    ),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
    select: ty.Optional[ty.List[str]] = None,
    as_json: bool = False,
) -> TaskDetail:
    """Print information about a task on the cluster"""
    res = resolve_task(name_or_id, broker_id=cluster_id, project_id=project_id)
    print_info_table(res, exclude=["plan"], as_json=as_json, select=select)
    return res


@cli.subcommand(
    "tasks",
    "logs",
    name_or_id=Arg(help=Messages.name_or_id_optional.format(noun="tasks")),
    project_id=Arg(help=Messages.project_id.format(noun="tasks")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="tasks")),
    as_json=Arg("--json", help=Messages.as_json),
)
def logs(
    name_or_id: ty.Optional[ty.StrOrUUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
    as_json: bool = False,
) -> ty.Optional[str]:
    """Get logs for a task on the cluster"""
    job = resolve_task(name_or_id, broker_id=cluster_id, project_id=project_id)
    auth = get_auth_state()
    if job.last_execution_id is None:
        raise CLIError(Messages.E051.format(noun="task", name=job.name, id=job.id))
    try:
        res = auth.broker_client.jobs.logs(
            JobLogRequest(id=job.last_execution_id, offset=0)
        )
    except BrokerError as e:
        raise CLIError(Messages.E011.format(noun="task"), e)
    text = res.text if res is not None else None
    print_logs(text, as_json=as_json)
    return text


@cli.subcommand(
    "tasks",
    "start",
    name_or_id=Arg(help=Messages.name_or_id_optional.format(noun="task")),
    project_id=Arg(help=Messages.project_id.format(noun="task")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="task")),
    worker_class=Arg(
        "--worker-class",
        help=Messages.recipe_worker_class.format(noun="task"),
    ),
)
def start(
    name_or_id: ty.Optional[ty.StrOrUUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
    worker_class: ty.Optional[str] = None,
) -> ty.UUID:
    """Start a task on the cluster"""
    job = resolve_task(name_or_id, broker_id=cluster_id, project_id=project_id)
    start_job(job, worker_class, get_auth_state())
    return job.id


@cli.subcommand(
    "tasks",
    "stop",
    name_or_id=Arg(help=Messages.name_or_id_optional.format(noun="task")),
    project_id=Arg(help=Messages.project_id.format(noun="task")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="task")),
)
def stop(
    name_or_id: ty.Optional[ty.StrOrUUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Stop a task on the cluster"""
    job = resolve_task(name_or_id, broker_id=cluster_id, project_id=project_id)
    stop_job(job, get_auth_state())
    return job.id


@cli.subcommand(
    "tasks",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="task")),
    project_id=Arg(help=Messages.project_id.format(noun="task")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="task")),
)
def delete(
    name_or_id: ty.StrOrUUID,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Delete a task by name or ID"""
    job = resolve_task(name_or_id, broker_id=cluster_id, project_id=project_id)
    delete_job(job, get_auth_state())
    return job.id



import ast
import itertools
import logging
import os
import platform
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
import warnings
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Literal, Optional, Set, Union

import packaging.tags
import packaging.utils
import pkginfo
import setuptools
from packaging.requirements import Requirement  # noqa: F401
from packaging.specifiers import SpecifierSet  # noqa: F401
from packaging.version import Version

from . import ty
from .appdirs import user_cache_dir

logger = logging.getLogger(__name__)

RecipeSourceKind = Literal["wheel", "distribution"]


class RecipeBuildVenvCreationFailed(Exception):
    pass


class RecipeSource(object):
    path: Path

    @classmethod
    def from_path(cls, path: Union[Path, str]) -> Optional["RecipeSource"]:
        if isinstance(path, str):
            path = Path(path)
        kind = cls._infer_kind(path)
        if kind == "wheel":
            return WheelSource(path)
        elif kind == "distribution":
            return DirectorySource(path)
        else:
            return None

    def __init__(
        self,
        path: Path,
    ):
        self.path = path

    @classmethod
    def _infer_kind(cls, path: Path) -> Optional[RecipeSourceKind]:
        if path.is_file() and path.suffix == ".whl":
            return "wheel"

        if path.is_dir() and (
            (path / "setup.py").is_file() or (path / "pyproject.toml").is_file()
        ):
            return "distribution"
        return None

    @cached_property
    def path_kind(self):
        return self._infer_kind(self.path)

    @cached_property
    def version(self) -> Optional[Version]:
        ...

    @cached_property
    def distribution_name(self) -> str:
        ...

    @cached_property
    def package_name(self) -> str:
        # TODO: this should be the name of the top level package, not the
        # distribution/wheel.
        assert self.distribution_name is not None

        package_name = re.sub(r"[^\w\d.]+", "_", self.distribution_name, re.UNICODE)
        return package_name

    @cached_property
    def pkginfo_metadata(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            maybe_pkginfo = pkginfo.get_metadata(str(self.path.resolve()))
            if maybe_pkginfo is not None and maybe_pkginfo.name is not None:
                return maybe_pkginfo
        return None

    @cached_property
    def requirements(self) -> Optional[Set[Requirement]]:
        if self.pkginfo_metadata is None:
            return None
        return set(Requirement(line) for line in self.pkginfo_metadata.requires_dist)


class WheelSource(RecipeSource):
    @cached_property
    def wheel_info(self) -> pkginfo.Wheel:
        metadata = self.pkginfo_metadata
        assert isinstance(
            metadata, pkginfo.Wheel
        ), f"expected pkginfo metadata to be Wheel, got {type(metadata)}"
        return metadata

    @cached_property
    def distribution_name(self) -> str:
        return self.wheel_info.name

    @cached_property
    def version(self) -> Version:
        return Version(self.wheel_info.version)

    def to_file_uri(self) -> str:
        return self.path.resolve().as_uri()


class DirectorySource(RecipeSource):
    def __init__(self, directory: Path):
        super().__init__(directory)
        # TODO: infer correctly
        name = self.package_name
        name_as_path = name.replace(".", os.sep)
        pkg_dir = directory / name_as_path
        src_pkg_dir = directory / "src" / name_as_path

        existing = set()
        if pkg_dir.is_dir():
            self.pkg_dir = pkg_dir
            self.prefix = ""
            existing.add(pkg_dir)
        if src_pkg_dir.is_dir():
            self.pkg_dir = src_pkg_dir
            self.prefix = "src"
            existing.add(src_pkg_dir)

        if len(existing) > 1:
            raise ValueError(
                "Multiple matches for package name {}: {}".format(
                    name, ", ".join([str(p) for p in sorted(existing)])
                )
            )
        elif not existing:
            raise ValueError("No package directory found for module {}".format(name))

        self.source_dir = directory / self.prefix
        self.directory = directory

        if "." in name:
            logger.warn(
                "Package name %s contains a dot and may not work as intended.", name
            )
            self.namespace_package_name = name.rpartition(".")[0]
            self.in_namespace_package = True

    @cached_property
    def version(self) -> Optional[Version]:
        from_ast = get_version_from_ast(self)
        if from_ast is not None:
            return Version(from_ast)
        # TODO: import in a subprocess or just build metadata eagerly using _build_meta
        # from_import = get_version_from_import(self)
        # if from_import is not None:
        #     return Version(from_import)
        return None

    @cached_property
    def distribution_name(self) -> str:
        cached_metadata = self.pkginfo_metadata
        if cached_metadata is not None:
            return cached_metadata.name
        # TODO: we probably need to do a wheel build to be sure
        distribution_name = re.sub(
            r"[^\w\d]+", "_", self.package_name, re.UNICODE
        ).replace("_", "-")
        return distribution_name

    @cached_property
    def discovered_packages(self):
        all_discovered = setuptools.find_packages(
            where=str(self.path), exclude=("tests", "tests.*")
        )
        return [p for p in all_discovered if p.find(".") == -1]

    @cached_property
    def package_name(self):
        discovered_packages = self.discovered_packages
        if len(discovered_packages) == 1:
            return discovered_packages[0]
        elif len(discovered_packages) == 0:
            raise ValueError(
                f"Could not infer package name from directory: {self.path}"
            )
        else:
            raise ValueError(
                f"Found multiple packages in directory {self.path}: {discovered_packages}"
            )

    @property
    def file(self):
        return self.pkg_dir / "__init__.py"

    @property
    def version_files(self):
        paths = [self.pkg_dir / "__init__.py"]
        for filename in ("about.py", "version.py", "_version.py", "__version__.py"):
            if (self.pkg_dir / filename).is_file():
                paths.insert(0, self.pkg_dir / filename)
        return paths

    def iter_files(self):
        def _include(path):
            name = os.path.basename(path)
            if (name == "__pycache__") or name.endswith(".pyc"):
                return False
            return True

        for dirpath, dirs, files in os.walk(str(self.pkg_dir)):
            for file in sorted(files):
                full_path = os.path.join(dirpath, file)
                if _include(full_path):
                    yield full_path

            dirs[:] = [d for d in sorted(dirs) if _include(d)]


def get_version_from_ast(target: DirectorySource):
    def _extract_version_assignment(node: ast.stmt) -> Optional[ast.Assign]:
        if not isinstance(node, ast.Assign):
            return None
        for target in node.targets:
            if isinstance(target, ast.Name) and target.id == "__version__":
                # TODO: multiple assignment?
                return node
        return None

    def _extract_literal_version(node: ast.Assign) -> Optional[str]:
        if isinstance(node.value, ast.Str):
            return node.value.s
        return None

    version = None
    node = None
    for version_path in target.version_files:
        with version_path.open("rb") as f:
            node = ast.parse(f.read())
        for child in node.body:
            assignment = _extract_version_assignment(child)
            if assignment is not None:
                version = _extract_literal_version(assignment)
                if version is not None:
                    break
            else:
                # programmatically generated version - we need to use imports
                pass
    assert node is not None, f"no __version__ found in module {target.package_name}"
    return version


_IMPORT_COUNTER = 0


def get_version_from_import(target: DirectorySource):
    """
    Import the module and extract the version from it.
    """
    global _IMPORT_COUNTER
    from importlib.util import module_from_spec, spec_from_file_location

    @contextmanager
    def stash_global_state():
        """
        Stash global state that modules might change at import time.

        TODO: Can we stash sys.modules, recipe registry, here?
        """
        logging_handlers = logging.root.handlers[:]
        try:
            yield
        finally:
            logging.root.handlers = logging_handlers

    _IMPORT_COUNTER += 1

    mod_name = f"prodigy_teams.imported_recipe.{_IMPORT_COUNTER}"
    spec = spec_from_file_location(mod_name, target.file)
    assert spec is not None, f"Failed to load module from {target.file}"
    assert spec.loader is not None
    with stash_global_state():
        m = module_from_spec(spec)
        sys.modules[mod_name] = m
        try:
            spec.loader.exec_module(m)
        finally:
            imported_mods = [mod for mod in sys.modules if mod.startswith(mod_name)]
            for mod in imported_mods:
                sys.modules.pop(mod, None)

    version = m.__dict__.get("__version__", None)
    return version


def get_version_from_pkginfo(target: DirectorySource):
    """
    Get the version from the .egg-info/.dist-info folders

    This version is computed when the package is built, so it's not
    necessarily the same as the version in the source code.
    """

    meta = pkginfo.get_metadata(str(target.source_dir))
    assert meta is not None
    return meta.version


class VenvError(Exception):
    pass


class InvalidVenvPath(VenvError):
    pass


class InvalidVenvError(VenvError):
    def __init__(
        self,
        venv_dir: Path,
        python_bin: ty.Optional[Path] = None,
        activate_script: ty.Optional[Path] = None,
    ):
        self.venv_dir = venv_dir
        self.python_bin = python_bin
        self.activate_script = activate_script
        super().__init__(
            f"Invalid venv: {venv_dir} ({python_bin=}, {activate_script=})"
        )


class PythonVersionError(VenvError):
    def __init__(
        self,
        venv_dir: Path,
        actual_version: str,
        required_version: str,
    ):
        self.venv_dir = venv_dir
        self.actual_version = actual_version
        self.required_version = required_version
        super().__init__(
            f"Invalid python version: {venv_dir} - version is {actual_version} but required {required_version}"
        )


class VenvCreationError(VenvError):
    def __init__(
        self, venv_dir: Path, build_venv: "Venv", error: subprocess.CalledProcessError
    ):
        self.venv_dir = venv_dir
        python_bin = Venv._find_python_bin(venv_dir)
        activate_script = Venv._find_activate(venv_dir)
        self.python_bin = python_bin
        self.activate_script = activate_script
        message = f"""Unexpected failure while creating virtualenv:
        ### Error: {error.__class__.__name__}
        cmd: {str(error.cmd)}
        return code: {str(error.returncode)}
        ### Stdout
        {error.stdout}
        ### Stderr
        {error.stderr}

        ### Debug Info:
        build_venv: {build_venv.python}
        - {Path(build_venv.python).exists()=}
        - {Path(build_venv.python).is_file()=}

        cached_venv: {str(venv_dir.absolute())}
        - {venv_dir=} ({"exists" if venv_dir and venv_dir.exists() else "not found"})
        - {python_bin=} ({"exists" if python_bin and python_bin.exists() else "not found"})
        - {activate_script=} ({"exists" if activate_script and activate_script.exists() else "not found"})
        """
        super().__init__(message)


class Venv:
    def __init__(
        self,
        python_binary: str,
        activate_script: ty.Optional[Path],
        venv_dir: ty.Optional[Path],
        cwd: ty.Optional[Path] = None,
    ) -> None:
        self.python = python_binary
        self.activate_script = activate_script
        self.venv_dir = venv_dir
        self._default_cwd = cwd

    @classmethod
    def init_from_dir(
        cls,
        venv_dir: Path,
        cwd: ty.Optional[Path],
        require_python_version: ty.Optional[str] = None,
    ) -> "Venv":
        python_bin, activate_script = cls._validate_venv(
            venv_dir, require_python_version=require_python_version
        )
        return Venv(str(python_bin), activate_script, venv_dir=venv_dir, cwd=cwd)

    @classmethod
    @contextmanager
    def from_active(cls, cwd: ty.Optional[Path]) -> ty.Iterator["Venv"]:
        yield cls(sys.executable, None, venv_dir=None, cwd=cwd)

    @classmethod
    @contextmanager
    def temporary(cls, cwd: ty.Optional[Path] = None) -> ty.Iterator["Venv"]:
        with _make_tempdir() as tmp:
            venv_path = Path(tmp) / "venv"
            with cls.from_active(cwd=cwd) as build_venv:
                yield build_venv.create_venv(venv_path, cwd=cwd)

    @classmethod
    def _get_venv_cache_path(cls, package_name: str, mkdir: bool = False) -> Path:
        cache_key = (package_name,)
        cache_base_path = _get_venvs_cache_path()
        cached_venv = (cache_base_path / "+".join(cache_key)).absolute()
        if mkdir:
            cached_venv.parent.mkdir(exist_ok=True, parents=True)
        return cached_venv

    def create_or_activate_venv(
        self,
        venv_path: Path,
        cwd: ty.Optional[Path] = None,
        fill_existing_dir: bool = False,
        require_python_version: ty.Optional[str] = None,
    ) -> "Venv":
        is_empty_dir = venv_path.is_dir() and len(os.listdir(venv_path)) == 0
        try_fill = fill_existing_dir and Venv._is_venv_dir(venv_path)
        if not venv_path.exists() or is_empty_dir or try_fill:
            with self.from_active(cwd=cwd) as build_venv:
                return build_venv.create_venv(venv_path, cwd=cwd)
        elif venv_path.is_dir():
            return Venv.init_from_dir(
                venv_path, cwd=cwd, require_python_version=require_python_version
            )
        else:
            raise InvalidVenvPath(f"Cannot create venv, path is a file: {venv_path}")

    def create_venv(self, venv_path: Path, cwd: ty.Optional[Path] = None) -> "Venv":
        try:
            _subprocess_run(
                [sys.executable, "-m", "venv", str(venv_path.absolute())],
                cwd or Path.cwd(),
            )
        except subprocess.CalledProcessError as e:
            raise VenvCreationError(venv_dir=venv_path, build_venv=self, error=e)
        return Venv.init_from_dir(venv_path, cwd=cwd)

    @classmethod
    def _is_venv_dir(cls, venv_path: Path) -> bool:
        if not venv_path.is_dir():
            return False
        python_bin = cls._find_python_bin(venv_path)
        activate_script = cls._find_activate(venv_path)
        return (
            python_bin is not None
            and python_bin.exists()
            or activate_script is not None
            and activate_script.exists()
        )

    @classmethod
    def _validate_venv(
        cls, venv_path: Path, require_python_version: ty.Optional[str] = None
    ) -> ty.Tuple[Path, Path]:
        python_bin = cls._find_python_bin(venv_path)
        activate_script = cls._find_activate(venv_path)
        if (
            python_bin is None
            or not python_bin.exists()
            or activate_script is None
            or not activate_script.exists()
        ):
            raise InvalidVenvError(
                venv_path, python_bin=python_bin, activate_script=activate_script
            )
        if require_python_version:
            try:
                output = _subprocess_run(
                    [str(python_bin.absolute()), "--version"], cwd=Path.cwd()
                ).stdout
                observed_version = output.split()[1]
                if not observed_version.startswith(require_python_version):
                    raise PythonVersionError(
                        venv_path,
                        actual_version=observed_version,
                        required_version=require_python_version,
                    )
            except Exception as e:
                raise InvalidVenvError(
                    venv_path, python_bin=python_bin, activate_script=activate_script
                ) from e

        return python_bin, activate_script

    @classmethod
    def _find_python_bin(cls, venv_base: Path) -> Optional[Path]:
        if sys.platform == "win32":
            return venv_base / "Scripts" / "python.exe"
        else:
            for candidate in ["local/bin", "bin"]:
                bin_path = venv_base / candidate / "python"
                if bin_path.exists():
                    return bin_path
        return None

    @classmethod
    def _find_activate(cls, venv_base: Path) -> Optional[Path]:
        if sys.platform == "win32":
            subdirs = ["Scripts"]
            bin_names = ["activate", "activate.bat"]
        else:
            subdirs = ["local/bin", "bin"]
            bin_names = ["activate", "activate.sh"]
        for subdir_candidate in subdirs:
            for bin_candidate in bin_names:
                bin_path = venv_base / subdir_candidate / bin_candidate
                if bin_path.exists():
                    return bin_path
        return None

    def cwd(self, override: ty.Optional[Path] = None) -> Path:
        return override or self._default_cwd or Path.cwd()

    def install(
        self,
        *packages: str,
        upgrade: bool = False,
        no_deps: bool = False,
        find_links: ty.List[Path] = [],
        cwd: ty.Optional[Path] = None,
        force_reinstall: bool = False,
    ) -> subprocess.CompletedProcess:
        args = [self.python, "-m", "pip", "install"]
        args.append("--disable-pip-version-check")
        if upgrade:
            args.append("--upgrade")
        if no_deps:
            args.append("--no-deps")
        if force_reinstall:
            args.append("--force-reinstall")
        if find_links:
            for dir in find_links:
                args.extend(["-f", str(dir)])
        args.extend(packages)
        return _subprocess_run(args, self.cwd(cwd), venv=self)

    def _run(
        self, *args: str, cwd: ty.Optional[Path] = None, input: ty.Optional[str] = None
    ) -> subprocess.CompletedProcess:
        return _subprocess_run([*args], self.cwd(cwd), venv=self, input=input)

    def run(
        self, *args: str, cwd: ty.Optional[Path] = None
    ) -> subprocess.CompletedProcess:
        return _subprocess_run([self.python, *args], self.cwd(cwd), venv=self)

    def run_module(
        self,
        entrypoint: str,
        *args: str,
        cwd: ty.Optional[Path] = None,
        input: ty.Optional[str] = None,
        env: ty.Optional[ty.Dict[str, str]] = None,
        print_command: bool = False,
        print_output: bool = False,
        popen_args: dict = {},
    ) -> subprocess.CompletedProcess:
        command = [self.python, "-m", entrypoint, *args]
        return _subprocess_run(
            command,
            self.cwd(cwd),
            input=input,
            env=env,
            venv=self,
            print_command=print_command,
            print_output=print_output,
            **popen_args,
        )

    @classmethod
    def cached(
        cls,
        package_name: str,
        build_venv: "Venv",
        cwd: ty.Optional[Path] = None,
        requirements: ty.Optional["RequirementSet"] = None,
    ) -> "Venv":
        venv_path = cls._get_venv_cache_path(package_name, mkdir=True)
        current_python_version = platform.python_version()
        try:
            venv = build_venv.create_or_activate_venv(
                venv_path, require_python_version=current_python_version
            )
        except (PythonVersionError, InvalidVenvError):
            shutil.rmtree(venv_path, ignore_errors=True)
            venv = build_venv.create_or_activate_venv(venv_path)

        if requirements:
            requirements_file = venv_path / "requirements.txt"
            cached_reqs = None
            if requirements_file.is_file():
                cached_reqs = RequirementSet.from_txt(requirements_file)
            if (
                cached_reqs is not None
                and cached_reqs != requirements
                or cached_reqs is None
            ):
                tmp_install = venv_path / "dependencies.txt"
                wheels = []
                # Split the installs into two parts:
                # 1. Non-wheel dependencies
                # 2. Local wheels (force reinstalled)
                tmp_install.write_text(
                    requirements.replace_local_wheel_links(
                        replace=lambda wheel_path: wheels.append(wheel_path)
                    ).to_txt()
                )
                # TODO:pip-sync with --python-executable pointing to venv python
                venv.install("-r", str(tmp_install.absolute()), no_deps=True)
                venv.install(
                    *[str(w.absolute()) for w in wheels],
                    no_deps=True,
                    force_reinstall=True,
                )
                requirements_file.write_text(requirements.to_txt())
                tmp_install.unlink()
        return venv


def wheel_file_to_version_pin(wheel_path: Path) -> Requirement:
    parsed = WheelName.from_filename(wheel_path.name)
    return Requirement(f"{parsed.name}=={str(parsed.version)}")


class RequirementSet:
    def __init__(self, requirements: ty.Iterable[ty.Union[Requirement, str]]) -> None:
        self._requirements = frozenset(
            [Requirement(r) if isinstance(r, str) else r for r in requirements]
        )

    def _validate(self, requirement: Requirement) -> None:
        if requirement.name is None:
            raise ValueError("Invalid requirement, must have a name")

    @cached_property
    def _by_name(self) -> ty.Dict[str, frozenset[Requirement]]:
        by_name = self.collect_by_name(self._requirements)
        return by_name

    def __iter__(self) -> ty.Iterator[Requirement]:
        return iter(sorted(self._requirements, key=lambda r: str(r)))

    def __len__(self) -> int:
        return len(self._requirements)

    def __contains__(self, item: ty.Any) -> bool:
        if isinstance(item, str):
            # item must be a distribution name
            return item in self._by_name
        elif isinstance(item, Requirement):
            # item is an exact requirement
            return item in self._requirements
        else:
            raise TypeError(f"Expected str or Requirement, got {type(item)}")

    def __getitem__(self, item: str) -> ty.Optional[frozenset[Requirement]]:
        return self._by_name.get(item, frozenset())

    def __repr__(self) -> str:
        return f"RequirementSet({list(self)})"

    def __eq__(self, other: ty.Any) -> bool:
        if not isinstance(other, RequirementSet):
            return NotImplemented
        return self._requirements == other._requirements

    def __hash__(self) -> int:
        return hash(self._requirements)

    @classmethod
    def collect_by_name(
        cls, *iterables: ty.Iterable[ty.Union[Requirement, str]]
    ) -> ty.Dict[str, frozenset[Requirement]]:
        by_name = defaultdict(set)
        for r in itertools.chain(*iterables):
            req = Requirement(r) if isinstance(r, str) else r
            by_name[req.name].add(req)
        return {k: frozenset(v) for k, v in by_name.items()}

    def partition(
        self, include: ty.Callable[[Requirement], bool]
    ) -> ty.Tuple["RequirementSet", "RequirementSet"]:
        included = []
        excluded = []
        for req in self:
            if include(req):
                included.append(req)
            else:
                excluded.append(req)
        return RequirementSet(included), RequirementSet(excluded)

    def merge(
        self, *others: ty.Iterable[ty.Union[str, Requirement]]
    ) -> "RequirementSet":
        return RequirementSet(itertools.chain(self, *others))

    def replace_local_wheel_links(
        self,
        wheels_to_replace: ty.Optional[ty.Set[str]] = None,
        replace: ty.Callable[
            [Path], ty.Optional[Requirement]
        ] = wheel_file_to_version_pin,
    ) -> "RequirementSet":
        def _should_replace(req: Requirement):
            name_matches = (
                wheels_to_replace is not None and req.name in wheels_to_replace
            )
            return (wheels_to_replace is None or name_matches) and _get_local_wheel_url(
                req
            ) is not None

        to_replace, keep = self.partition(_should_replace)
        replaced = []
        for req in to_replace:
            wheel_path = _get_local_wheel_url(req)
            assert (
                wheel_path is not None
            ), "expected local wheel to have a file:// url in compiled requirements file"
            replacement = replace(wheel_path)
            if replacement is not None:
                replaced.append(replacement)
        return keep.merge(replaced)

    @classmethod
    def from_txt(
        cls, txt: ty.Union[str, Path], ignore_comments: bool = True
    ) -> "RequirementSet":
        if isinstance(txt, Path):
            txt = txt.read_text()
        lines = txt.splitlines()
        lines = (line for line in lines if line.strip())
        if ignore_comments:
            lines = (line for line in lines if not line.lstrip().startswith("#"))

        return RequirementSet(lines)

    def to_txt(self) -> str:
        return "\n".join(str(req) for req in self)

    def to_lines(self) -> ty.List[str]:
        return [str(req) for req in self]


@dataclass
class WheelName:
    filename: str
    name: packaging.utils.NormalizedName
    version: str
    build: ty.Optional[ty.Tuple[int, str]]
    tags: ty.Set[packaging.tags.Tag]

    @classmethod
    def from_filename(cls, filename: str) -> "WheelName":
        name, version, build, tags = packaging.utils.parse_wheel_filename(filename)
        return cls(
            filename=filename,
            name=name,
            version=str(version),
            build=build if build else None,
            tags=set(tags),
        )


def _get_venvs_cache_path():
    cache_path = user_cache_dir(
        # the version key here should be updated if we make breaking changes to the cache contents
        appname="Prodigy Teams",
        appauthor="Explosion",
        version="0.1",
    )
    return Path(cache_path) / "venvs"


def _get_local_wheel_url(req: Requirement) -> ty.Optional[Path]:
    if req.url is not None and req.url.startswith("file://"):
        url_path = Path(req.url.replace("file://", ""))
        if url_path.suffix == ".whl":
            return url_path
    return None


def _utf8_decode(b: bytes) -> str:
    return b.decode()


def _source_sh(
    activate_path: Path,
    *,
    env: ty.Optional[ty.Dict[str, str]] = None,
    cwd: Path,
    decode_key=_utf8_decode,
    decode_value=_utf8_decode,
) -> ty.Dict[str, str]:
    cmd = f". {shlex.quote(str(activate_path.absolute()))}; env -0"
    ret = subprocess.run(
        cmd,
        shell=True,
        check=True,
        capture_output=True,
        env=env,
    )
    result = {}
    kvs = [line.partition(b"=")[::2] for line in ret.stdout.split(b"\0")]
    result.update((decode_key(k), decode_value(v)) for k, v in kvs)
    return result


def _subprocess_run(
    args: ty.List[str],
    cwd: Path,
    input: ty.Optional[str] = None,
    env: ty.Optional[ty.Dict[str, str]] = None,
    venv: ty.Optional[Venv] = None,
    print_command: bool = False,
    print_output: bool = False,
    popen_args: dict = {},
) -> subprocess.CompletedProcess:
    if env is None:
        env = os.environ.copy()

    if venv and venv.activate_script is not None:
        venv_config = _source_sh(venv.activate_script, cwd=cwd, env=env)
        for k in ["PATH", "VIRTUAL_ENV", "PYTHONHOME", "VIRTUAL_ENV_PROMPT"]:
            if k in venv_config:
                env[k] = venv_config[k]
            elif k in env:
                del env[k]

    if print_command:
        print(shlex.join(args), file=sys.stderr)
    try:
        ret = subprocess.run(
            args,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=cwd,
            text=True,
            input=input,
            encoding="utf8",
            env=env,
            **popen_args,
        )
    except subprocess.CalledProcessError as e:
        print(e.output, file=sys.stderr)
        raise
    if print_output:
        print(ret.stdout, file=sys.stderr)
    return ret


@contextmanager
def _make_tempdir() -> ty.Generator[Path, None, None]:
    """Execute a block in a temporary directory and remove the directory and
    its contents at the end of the with block.

    YIELDS (Path): The path of the temp directory.
    """
    d = Path(tempfile.mkdtemp())
    yield d
    shutil.rmtree(str(d))



from ..models import (
    OrgCreating,
    OrgDeleting,
    OrgDetail,
    OrgReading,
    OrgSummary,
    OrgUpdating,
)
from .base import ModelClient


class Org(
    ModelClient[
        OrgCreating, OrgReading, OrgUpdating, OrgDeleting, OrgSummary, OrgDetail
    ]
):
    Creating = OrgCreating
    Reading = OrgReading
    Updating = OrgUpdating
    Deleting = OrgDeleting
    Summary = OrgSummary
    Detail = OrgDetail



from ..models import (
    DatasetCreating,
    DatasetDeleting,
    DatasetDetail,
    DatasetReading,
    DatasetSummary,
    DatasetUpdating,
)
from .base import ModelClient


class Dataset(
    ModelClient[
        DatasetCreating,
        DatasetReading,
        DatasetUpdating,
        DatasetDeleting,
        DatasetSummary,
        DatasetDetail,
    ]
):
    Creating = DatasetCreating
    Reading = DatasetReading
    Updating = DatasetUpdating
    Deleting = DatasetDeleting
    Summary = DatasetSummary
    Detail = DatasetDetail



from .. import ty
from ..models import (
    JobActionResponse,
    JobID,
    JobLogRequest,
    JobLogResponse,
    TasksBaseUrl,
)
from .base import BaseClient


class Jobs(BaseClient):
    def get_base_url(self) -> TasksBaseUrl:
        res = self.request("POST", endpoint="get-base-url", return_model=TasksBaseUrl)
        return ty.cast(TasksBaseUrl, res)

    async def get_base_url_async(self) -> TasksBaseUrl:
        res = await self.request_async(
            "POST", endpoint="get-base-url", return_model=TasksBaseUrl
        )
        return ty.cast(TasksBaseUrl, res)

    def start_job(self, body: JobID) -> JobActionResponse:
        res = self.request(
            "POST", endpoint="start-job", data=body, return_model=JobActionResponse
        )
        return ty.cast(JobActionResponse, res)

    async def start_job_async(self, body: JobID) -> JobActionResponse:
        res = await self.request_async(
            "POST", endpoint="start-job", data=body, return_model=JobActionResponse
        )
        return ty.cast(JobActionResponse, res)

    def stop_job(self, body: JobID) -> JobActionResponse:
        res = self.request(
            "POST", endpoint="stop-job", data=body, return_model=JobActionResponse
        )
        return ty.cast(JobActionResponse, res)

    async def stop_job_async(self, body: JobID) -> JobActionResponse:
        res = await self.request_async(
            "POST", endpoint="stop-job", data=body, return_model=JobActionResponse
        )
        return ty.cast(JobActionResponse, res)

    def delete_job(self, body: JobID) -> JobActionResponse:
        res = self.request(
            "POST", endpoint="delete-job", data=body, return_model=JobActionResponse
        )
        return ty.cast(JobActionResponse, res)

    async def delete_job_async(self, body: JobID) -> JobActionResponse:
        res = await self.request_async(
            "POST", endpoint="delete-job", data=body, return_model=JobActionResponse
        )
        return ty.cast(JobActionResponse, res)

    def logs(self, body: JobLogRequest) -> JobLogResponse:
        res = self.request(
            "POST", endpoint="logs", data=body, return_model=JobLogResponse
        )
        return ty.cast(JobLogResponse, res)

    async def logs_async(self, body: JobLogRequest) -> JobLogResponse:
        res = await self.request_async(
            "POST", endpoint="logs", data=body, return_model=JobLogResponse
        )
        return ty.cast(JobLogResponse, res)



import httpx

from .. import ty
from ..models import (
    Copying,
    CopyPlan,
    Deleting,
    Downloading,
    FileStats,
    Listing,
    PathList,
    RsyncPlan,
    Statting,
)
from .base import BaseClient


class Files(BaseClient):
    def upload(
        self, file: ty.IO, *, dest: str, overwrite: bool, make_dirs: bool
    ) -> None:
        res = self.request(
            "POST",
            endpoint="upload",
            files={"file": file},
            params={"dest": dest, "overwrite": overwrite, "make_dirs": make_dirs},
            return_model=None,
        )
        return ty.cast(None, res)

    async def upload_async(
        self, file: ty.IO, *, dest: str, overwrite: bool, make_dirs: bool
    ) -> None:
        res = await self.request_async(
            "POST",
            endpoint="upload",
            files={"file": file},
            params={"dest": dest, "overwrite": overwrite, "make_dirs": make_dirs},
            return_model=None,
        )
        return ty.cast(None, res)

    def copy(self, body: Copying) -> None:
        res = self.request("POST", endpoint="copy", data=body, return_model=None)
        return ty.cast(None, res)

    async def copy_async(self, body: Copying) -> None:
        res = await self.request_async(
            "POST", endpoint="copy", data=body, return_model=None
        )
        return ty.cast(None, res)

    def delete(self, body: Deleting) -> None:
        res = self.request("POST", endpoint="delete", data=body, return_model=None)
        return ty.cast(None, res)

    async def delete_async(self, body: Deleting) -> None:
        res = await self.request_async(
            "POST", endpoint="delete", data=body, return_model=None
        )
        return ty.cast(None, res)

    def stat(self, body: Statting) -> FileStats:
        res = self.request("POST", endpoint="stat", data=body, return_model=FileStats)
        return ty.cast(FileStats, res)

    async def stat_async(self, body: Statting) -> FileStats:
        res = await self.request_async(
            "POST", endpoint="stat", data=body, return_model=FileStats
        )
        return ty.cast(FileStats, res)

    def list_dir(self, body: Listing) -> PathList:
        res = self.request(
            "POST",
            endpoint="list-dir",
            data=body,
            return_model=PathList,
            timeout=httpx.Timeout(5.0, read=20.0),
        )
        return ty.cast(PathList, res)

    async def list_dir_async(self, body: Listing) -> PathList:
        res = await self.request_async(
            "POST",
            endpoint="list-dir",
            data=body,
            return_model=PathList,
            timeout=httpx.Timeout(5.0, read=20.0),
        )
        return ty.cast(PathList, res)

    def plan_directory_copy(self, body: Copying) -> CopyPlan:
        res = self.request(
            "POST", endpoint="plan-directory-copy", data=body, return_model=CopyPlan
        )
        return ty.cast(CopyPlan, res)

    async def plan_directory_copy_async(self, body: Copying) -> CopyPlan:
        res = await self.request_async(
            "POST", endpoint="plan-directory-copy", data=body, return_model=CopyPlan
        )
        return ty.cast(CopyPlan, res)

    def plan_directory_rsync(self, body: Copying) -> RsyncPlan:
        res = self.request(
            "POST", endpoint="plan-directory-rsync", data=body, return_model=CopyPlan
        )
        return ty.cast(RsyncPlan, res)

    async def plan_directory_rsync_async(self, body: Copying) -> RsyncPlan:
        res = await self.request_async(
            "POST", endpoint="plan-directory-rsync", data=body, return_model=CopyPlan
        )
        return ty.cast(RsyncPlan, res)

    def download(self, body: Downloading) -> ty.IO:  # StreamingResponse
        res = self.request("POST", endpoint="download", data=body, stream=True)
        return ty.cast(ty.IO, res)

    async def download_async(self, body: Downloading) -> ty.IO:
        res = await self.request_async(
            "POST", endpoint="download", data=body, stream=True
        )
        return ty.cast(ty.IO, res)



from ..models import (
    ActionCreating,
    ActionDeleting,
    ActionDetail,
    ActionReading,
    ActionSummary,
    ActionUpdating,
)
from .base import ModelClient


class Action(
    ModelClient[
        ActionCreating,
        ActionReading,
        ActionUpdating,
        ActionDeleting,
        ActionSummary,
        ActionDetail,
    ]
):
    Creating = ActionCreating
    Reading = ActionReading
    Updating = ActionUpdating
    Deleting = ActionDeleting
    Summary = ActionSummary
    Detail = ActionDetail



from . import ty


class BrokerError(Exception):
    def __init__(
        self,
        message: ty.Optional[str] = None,
        detail: ty.Optional[str] = None,
        status: ty.Optional[int] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__()
        self.message = message
        self.detail = detail
        self.status = status
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        output = ""
        if self.message:
            output = self.message
        if self.detail:
            output += f"\n{self.detail}" if output else self.detail
        return output


class NotFound(BrokerError):
    pass


class Exists(BrokerError):
    pass


class ClientError(BrokerError):
    pass


class PT_MaintenanceMode(BrokerError):
    pass


class DatasetCreateFailed(BrokerError):
    pass


class DatasetExists(BrokerError):
    pass


class DatasetNotFound(BrokerError):
    pass


class PT_Package_Exists(BrokerError):
    pass


class PT_Package_NotFound(BrokerError):
    pass


class PackageNotWheel(BrokerError):
    pass


class EnvNotFound(BrokerError):
    pass


class JobMetaNotFound(BrokerError):
    pass


class FileError(BrokerError):
    pass


class RmPathNotFoundError(FileError):
    pass


class RmPathIsADirectoryError(FileError):
    pass


class SecretCreateFailed(BrokerError):
    pass


class SecretExists(BrokerError):
    pass


class SecretNotFound(BrokerError):
    pass


class AuthError(BrokerError):
    def __init__(self, detail: str = "Auth Error"):
        self.status = 401
        self.detail = detail

    def __str__(self):
        return self.detail



import json
import os
import re
import shutil
import sys
import textwrap

from radicli import HelpFormatter
from radicli.util import format_arg_help
from wasabi import color, msg, table

from . import ty
from .about import __prog__
from .messages import Messages
from .prodigy_teams_pam_sdk.models import RecipeDetail
from .util import CYGWIN, WIN

LOGO = """
█▀█ █▀█ █▀█ █▀▄ █ █▀▀ █▄█   ▀█▀ █▀▀ ▄▀█ █▀▄▀█ █▀
█▀▀ █▀▄ █▄█ █▄▀ █ █▄█ ░█░   ░█░ ██▄ █▀█ █░▀░█ ▄█
"""

whitespace_re = re.compile(r"\s+", re.ASCII)
ansi_re = re.compile(r"\033\[[;?0-9]*[a-zA-Z]")


def is_silent() -> bool:
    # External environment variable to disable printing of info tables etc.
    return bool(os.environ.get("PRODIGY_TEAMS_CLI_SILENT", False))


def print_login_info(uri_complete: str, uri: str, code: str) -> None:
    print(LOGO)
    msg.info(Messages.T014)
    open_url(uri_complete)
    msg.text(Messages.T015)
    msg.text(color(uri_complete, bold=True))
    print("")
    msg.text(Messages.T016)
    msg.text(color(uri, bold=True))
    msg.text(Messages.T017)
    msg.text(color(code, bold=True))


def print_as_json(data: ty.JSONableDict) -> None:
    if not is_silent():
        print(json.dumps(data, indent=2, default=str))


def print_logs(text: ty.Optional[str], as_json: bool = False) -> None:
    if is_silent():
        return
    if as_json:
        print_as_json({"logs": text})
    elif text is not None:
        msg.info("Logs")
        print(text)


def dicts_to_table(
    data: ty.Sequence[ty.Dict[str, ty.Any]],
    headers: ty.Optional[ty.List[str]] = None,
    exclude_dicts: bool = False,
) -> ty.Tuple[
    ty.List[str], ty.Union[ty.Sequence[ty.Dict[str, ty.Any]], ty.List[ty.List[str]]]
]:
    """
    Generate a tuple with headers and rows in the format wasabi expects from a list
    of dicts having column names as keys.
    """
    if not data:
        return [], data
    use_data = list(data)
    if headers is None:
        headers = list(use_data[0].keys())
    rows = []
    for item in data:
        if exclude_dicts:
            item = {k: v for k, v in item.items() if not isinstance(v, dict)}
        row = []
        for key in headers:
            val = item.get(key, None)
            if isinstance(val, ty.Enum):
                val = val.value
            row.append(str(val))
        rows.append(row)
    return headers, rows


def print_table_with_select(
    items: ty.Sequence[ty.Union[ty.BaseModel, dict]],
    select: ty.Optional[ty.List[str]] = None,
    as_json: bool = False,
) -> None:
    if is_silent():
        return
    include: ty.Set[str] = set(select) if select is not None else None  # type: ignore

    def to_dict(r: ty.Any) -> ty.JSONableDict:
        if isinstance(r, ty.BaseModel):
            d = r.dict(include=include)
        elif hasattr(r, "as_dict"):
            d = r.as_dict(include=include)
        else:
            d = r
        if include:
            return {k: v for k, v in d.items() if k in include}
        return d

    if as_json:
        print(json.dumps([to_dict(item) for item in items], default=str))
    else:
        data = [to_dict(r) for r in items]
        headers, rows = dicts_to_table(data, headers=select)
        if rows:
            msg.table(rows, header=headers, divider=True, max_col=3000)


def print_info_table(
    item: ty.BaseModel,
    select: ty.Optional[ty.List[str]] = None,
    exclude: ty.Optional[ty.List[str]] = None,
    as_json: bool = False,
) -> None:
    if is_silent():
        return
    dict_kwargs = {}
    if select is not None:
        dict_kwargs["include"] = set(select)
    if exclude is not None:
        dict_kwargs["exclude"] = set(exclude)
    if as_json:
        print(item.json(**dict_kwargs, indent=2))
    else:
        serialized = {}
        for key, val in item._iter(**dict_kwargs, to_dict=True):
            if isinstance(val, ty.Enum):
                serialized[key] = val.value
            else:
                serialized[key] = val
        msg.table(serialized)


def print_args_table(
    args: ty.Dict[str, ty.Any], cli_names: ty.Optional[ty.Dict[str, str]] = None
) -> None:
    if is_silent():
        return
    data = {}
    cli_names_map = {v: k for k, v in cli_names.items()} if cli_names else {}
    for key, value in args.items():
        key = cli_names_map.get(key, key)
        value = str(value) if isinstance(value, ty.UUID) else repr(value)
        data[key] = value
    msg.table(data)


def print_recipes_help(
    schemas: ty.Dict[str, RecipeDetail], help: str, command: str
) -> None:
    data = []
    for name, schema in schemas.items():
        data.append((name, schema.title))
        if schema.description:
            data.append(("", format_arg_help(schema.description)))
    print(f"usage: {__prog__} {command} [-h] name [...]\n")
    print(
        f"{help}\nRun {__prog__} {command} [name] --help to view the recipe arguments."
    )
    print("")
    msg.text("Available recipes:")
    msg.table(data)


def get_recipe_desc(schema: RecipeDetail) -> str:
    width = max(shutil.get_terminal_size().columns - 2, 11)
    text = whitespace_re.sub(" ", schema.description or "").strip()
    desc = textwrap.fill(text, width, initial_indent="", subsequent_indent="")
    meta = {
        "ID": schema.id,
        "Source": schema.entry_point,
        "Created": schema.created.strftime("%Y-%m-%d, %H:%M:%S"),
    }
    meta_table = table(meta)
    return f"{schema.title} ({schema.name})\n{desc}\n{meta_table}"


class RecipeHelpFormatter(HelpFormatter):
    def _fill_text(self, text, width, indent) -> str:
        return "".join(indent + line for line in text.splitlines(keepends=True))


# Source: https://github.com/pallets/click/blob/cba52fa76135af2edf46c154203b47106f898eb3/src/click/_termui_impl.py#L551
def open_url(url: str, wait: bool = False, locate: bool = False) -> int:
    import subprocess

    def _unquote_file(url: str) -> str:
        from urllib.parse import unquote

        if url.startswith("file://"):
            url = unquote(url[7:])

        return url

    if sys.platform == "darwin":
        args = ["open"]
        if wait:
            args.append("-W")
        if locate:
            args.append("-R")
        args.append(_unquote_file(url))
        null = open("/dev/null", "w")
        try:
            return subprocess.Popen(args, stderr=null).wait()
        finally:
            null.close()
    elif WIN:
        if locate:
            url = _unquote_file(url.replace('"', ""))
            args = f'explorer /select,"{url}"'
        else:
            url = url.replace('"', "")
            wait_str = "/WAIT" if wait else ""
            args = f'start {wait_str} "" "{url}"'
        return os.system(args)
    elif CYGWIN:
        if locate:
            url = os.path.dirname(_unquote_file(url).replace('"', ""))
            args = f'cygstart "{url}"'
        else:
            url = url.replace('"', "")
            wait_str = "-w" if wait else ""
            args = f'cygstart {wait_str} "{url}"'
        return os.system(args)
    try:
        if locate:
            url = os.path.dirname(_unquote_file(url)) or "."
        else:
            url = _unquote_file(url)
        c = subprocess.Popen(["xdg-open", url])
        if wait:
            return c.wait()
        return 0
    except OSError:
        if url.startswith(("http://", "https://")) and not locate and not wait:
            import webbrowser

            webbrowser.open(url)
            return 0
        return 1


# Adapted from: https://github.com/pallets/click/blob/cba52fa76135af2edf46c154203b47106f898eb3/src/click/_termui_impl.py#L35
# Only supports sized sequences and basic percent / count
_BarT = ty.TypeVar("_BarT")


class Progress(ty.Generic[_BarT]):
    def __init__(
        self,
        iterable: ty.Union[ty.Iterable[_BarT], ty.Sequence[_BarT]],
        length: ty.Optional[int] = None,
        pc: bool = False,
        width: int = 36,
    ) -> None:
        self.show_pc = pc
        self.file = sys.stdout
        self._completed_intervals = 0
        self.width = width
        self.before_bar = "\r" if os.name == "nt" else "\r\033[?25l"
        self.after_bar = "\n" if os.name == "nt" else "\033[?25h\n"
        self.iter = iter(iterable)
        self.length = (
            length
            if length is not None
            else len(iterable)
            if isinstance(iterable, ty.Sequence)
            else 0
        )
        self.pos = 0
        self.finished = False
        self.max_width: ty.Optional[int] = None
        self.current_item: ty.Optional[_BarT] = None
        self.is_hidden = not isatty(self.file)
        self._last_line: ty.Optional[str] = None

    def __enter__(self) -> "Progress[_BarT]":
        self.render_progress()
        return self

    def __exit__(self, exc_type, exc_value, tb):  # type: ignore
        self.render_finish()

    def __iter__(self) -> ty.Iterator[_BarT]:
        self.render_progress()
        return self.generator()

    def __next__(self) -> _BarT:
        return next(iter(self))

    def render_finish(self) -> None:
        if self.is_hidden:
            return
        self.file.write(self.after_bar)
        self.file.flush()

    @property
    def pct(self) -> float:
        if self.length == 0:
            return 1.0
        return 1.0 if self.finished else min(self.pos / float(self.length), 1.0)

    def format_bar(self) -> str:
        empty_char = " "
        fill_char = "#"
        if self.finished:
            return fill_char * self.width
        bar_length = int(self.pct * self.width)
        bar = fill_char * bar_length
        bar += empty_char * (self.width - bar_length)
        return bar

    def render_progress(self) -> None:
        if self.is_hidden:
            return
        buf = []
        clear_width = self.width
        if self.max_width is not None:
            clear_width = self.max_width
        buf.append(self.before_bar)
        info = (
            f"{int(self.pct * 100): 4}%"[1:]
            if self.show_pc
            else f"{str(self.pos)}/{self.length}"
        )
        line = f"[{self.format_bar()}]  {info}".rstrip()
        line_len = len(ansi_re.sub("", line))
        if self.max_width is None or self.max_width < line_len:
            self.max_width = line_len
        buf.append(line)
        buf.append(" " * (clear_width - line_len))
        line = "".join(buf)
        if line != self._last_line:
            self._last_line = line
            self.file.write(line)
            self.file.flush()

    def make_step(self, n_steps: int) -> None:
        self.pos += n_steps
        if self.pos >= self.length:
            self.finished = True

    def update(self, n_steps: int, current_item: ty.Optional[_BarT] = None) -> None:
        if current_item is not None:
            self.current_item = current_item
        self._completed_intervals += n_steps
        if self._completed_intervals >= 1:
            self.make_step(self._completed_intervals)
            self.render_progress()
            self._completed_intervals = 0

    def finish(self) -> None:
        self.current_item = None
        self.finished = True

    def generator(self) -> ty.Iterator[_BarT]:
        if self.is_hidden:
            yield from self.iter
        else:
            for rv in self.iter:
                self.current_item = rv
                if self._completed_intervals == 0:
                    self.render_progress()
                yield rv
                self.update(1)
            self.finish()
            self.render_progress()


def isatty(stream: ty.IO) -> bool:
    try:
        return stream.isatty()
    except Exception:
        return False


class PageProgress(Progress[ty.Page]):
    def __init__(self, iterable: ty.Iterable[ty.Page], *args, length=1, **kwargs):
        super().__init__(iterable, *args, length=length, **kwargs)

        self._pos = 0

    def generator(self) -> ty.Iterator[ty.Page]:
        for page in super().generator():
            self.length = page.total
            self._pos += len(page.items)
            self.pos = self._pos - 1

            yield page



import functools
import posixpath
from pathlib import Path, PurePosixPath

from radicli import Arg
from wasabi import msg

from ... import ty
from ...cli import cli
from ...errors import BrokerError, CLIError
from ...messages import Messages
from ...prodigy_teams_broker_sdk import Client as BrokerClient
from ...prodigy_teams_broker_sdk import models as broker_models
from ...ui import Progress, is_silent
from ...util import _is_root_path, _resolve_broker_ref, is_local, resolve_remote_path
from .._state import get_auth_state, get_saved_settings


class CopyCallable(ty.Protocol):
    def __call__(
        self,
        *,
        src: str,
        dest: str,
        overwrite: bool,
        recurse: bool,
        make_dirs: bool,
        broker_client: BrokerClient,
    ) -> ty.Optional[str]:
        ...


@cli.subcommand(
    "files",
    "cp",
    src=Arg(help=Messages.remote_local_path.format(noun="source")),
    dest=Arg(help=Messages.remote_local_path.format(noun="destination")),
    recurse=Arg("--recurse", "-r", help=Messages.recurse_copy),
    make_dirs=Arg("--make-dirs", help=Messages.make_dirs),
    overwrite=Arg("--overwrite", help=Messages.overwrite),
    cluster_host=Arg("--cluster-host", help=Messages.cluster_host),
)
def cp(
    src: str,
    dest: str,
    recurse: bool = False,
    make_dirs: bool = False,
    overwrite: bool = False,
    cluster_host: ty.Optional[str] = None,
) -> None:
    """Copy files to and from the cluster"""
    settings = get_saved_settings()
    auth = get_auth_state()
    broker_host = str(
        _resolve_broker_ref(auth.client, cluster_host or settings.broker_host)
    )
    is_src_local = is_local(src)
    is_dest_local = is_local(dest)

    def _cp(f: CopyCallable, *, src: str, dest: str) -> ty.Optional[str]:
        return f(
            src=src,
            dest=dest,
            overwrite=overwrite,
            recurse=recurse,
            make_dirs=make_dirs,
            broker_client=auth.broker_client,
        )

    if is_src_local and is_dest_local:  # both local
        raise CLIError(Messages.E015.format(verb="copy"), f"{src}, {dest}")
    elif is_src_local and not is_dest_local:  # from local to remote
        err = _cp(
            _upload,
            src=src,
            dest=resolve_remote_path(auth.client, dest, broker_host),
        )
    elif not is_src_local and is_dest_local:
        # From remote to local
        err = _cp(
            _download,
            src=resolve_remote_path(auth.client, src, broker_host),
            dest=dest,
        )
    else:
        # Both remote
        err = _cp(
            _copy_on_remote,
            src=resolve_remote_path(auth.client, src, broker_host),
            dest=resolve_remote_path(auth.client, dest, broker_host),
        )
    if err is not None:
        raise CLIError(err)
    elif not is_silent():
        msg.good(Messages.T011, f"{src} -> {dest}")


def _upload(
    *,
    src: str,
    dest: str,
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
) -> ty.Optional[str]:
    src_has_trailing_slash = src.endswith("/")
    src_path = Path(src)
    src_base = src

    @functools.lru_cache()
    def _ls(remote_path: str):
        try:
            files = broker_client.files.list_dir(
                broker_models.Listing(
                    path=remote_path, recurse=False, include_stats=False
                )
            )
            return files
        except BrokerError as e:
            raise CLIError(Messages.E018, e)

    src_exists = src_path.exists()
    src_is_file = src_path.is_file()
    src_is_dir = src_path.is_dir()

    dest_exists = _ls(dest).exists
    dest_is_file = _ls(dest).is_file
    dest_is_dir = dest_exists and not dest_is_file

    # print(
    #     f"""
    # _upload({src=},{dest=}):
    #            {_ls(dest).paths=}
    #            {src_is_file=}
    #            {src_is_dir=}
    #            {src_has_trailing_slash=}
    #
    #            {dest_exists=}
    #            {dest_is_dir=}
    #            {dest_is_file=}
    # """.strip(),
    #     file=sys.stderr,
    # )

    if not src_exists:
        raise CLIError("No such file or directory", src)

    if src_is_dir and not recurse:
        raise CLIError(Messages.E016.format(verb="upload"), src_path)

    if dest_is_file:
        if src_is_file and not overwrite:
            raise CLIError("The destination file already exists and overwrite is False")
        elif src_is_dir:
            raise CLIError("A file already exists at the destination path")

    if src_is_file:
        src_file = src_path
        if dest_is_file:
            dest_file = dest
        elif dest_is_dir or dest.endswith("/"):
            dest_file = posixpath.join(dest, src_path.name)
        else:
            dest_file = dest
        dest_base = posixpath.dirname(dest_file)
        create_base = False
        paths = [(src_file, dest_file)]
    else:
        if dest_is_dir and src_has_trailing_slash:
            # `cp src/ dest_exists/` copies `src/**` into `dest_exists/**`
            dest_base = dest
            create_base = False
        elif dest_is_dir and not src_has_trailing_slash:
            # `cp src dest_exists/` copies `src/**` into `dest_exists/src/**`
            dest_base = posixpath.join(dest, src_path.name)
            create_base = True
        else:
            # `cp src dest_noexists/` copies `src/**` into `dest_noexists/**`
            create_base = True
            dest_base = dest
        src_base = src_path

        paths = [
            (
                Path(src_file),
                posixpath.join(
                    dest_base, str(PurePosixPath(src_file.relative_to(src_base)))
                ),
            )
            for src_file in src_path.glob("**/*")
        ]

    if not make_dirs and not _is_root_path(dest_base):
        # we can't precreate empty dirs on cloud paths, but we can
        # override the make_dirs flag to create the appropriate dirs on demand
        # if we first check the existence of the parent dirs
        required_parent = posixpath.dirname(dest_base) if create_base else dest_base

        parent_ls = _ls(required_parent)

        # print(
        #     f"""
        #        {required_parent=}
        #        {parent_ls.exists=}
        #        {parent_ls.is_file=}
        #        {create_base=}
        # """.rstrip(),
        #     file=sys.stderr,
        # )
        if parent_ls.exists and not parent_ls.is_file:
            make_dirs = True
        elif parent_ls.is_file:
            raise CLIError(
                "A file already exists at the destination path", required_parent
            )
        else:
            raise CLIError(Messages.E021, dest_base)

    #     print(
    #         f"""
    #                {paths=}
    # """.rstrip(),
    #         file=sys.stderr,
    #     )
    with Progress(paths) as paths:
        for src_file, dest_file in paths:
            if src_file.is_dir():
                continue
            body = src_file.open("rb")
            try:
                broker_client.files.upload(
                    body, dest=dest_file, overwrite=overwrite, make_dirs=make_dirs
                )
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="upload"), e)


def _download(
    *,
    src: str,
    dest: ty.Union[str, Path],
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
    verbose: bool = False,
) -> ty.Optional[str]:
    dest = Path(dest)
    body = broker_models.Listing(path=src, recurse=recurse, include_stats=False)
    try:
        files = broker_client.files.list_dir(body)
    except BrokerError as e:
        raise CLIError(Messages.E018, e)

    src_exists = files.exists
    src_is_file = files.is_file
    src_is_dir = src_exists and not src_is_file
    src_has_trailing_slash = src.endswith("/")

    dest_is_dir = dest.is_dir()
    dest_is_file = dest.is_file()

    #     print(
    #         f"""
    # _download({src=},{dest=}):
    #            {files.paths=}
    #            {src_is_file=}
    #            {src_is_dir=}
    #            {src_has_trailing_slash=}
    #
    #            {dest_exists=}
    #            {dest_is_dir=}
    #            {dest_is_file=}
    # """.strip()
    #     file=sys.stderr,
    #     )

    if not src_exists and not files.paths:
        raise CLIError("No such file or directory", src)

    if dest_is_file:
        if src_is_file and not overwrite:
            raise CLIError(Messages.E020, dest)
        elif src_is_dir:
            raise CLIError("A file already exists at the destination path")

    if src_is_dir and not recurse:
        raise CLIError(Messages.E016.format(verb="download"), src)

    if src_is_file:
        assert len(files.paths) == 1
        src_file = files.paths[0]
        if dest_is_file:
            dest_file = dest
        elif dest_is_dir:
            dest_file = dest / posixpath.basename(src_file)
        else:
            dest_file = dest
        dest_base = dest_file.parent
        create_base = False
        paths = [(src_file, dest_file)]
    else:
        if dest_is_dir and src_has_trailing_slash:
            # `cp src/ dest_exists/` copies `src/**` into `dest_exists/**`
            dest_base = dest
            create_base = False
        elif dest_is_dir and not src_has_trailing_slash:
            # `cp src dest_exists/` copies `src/**` into `dest_exists/src/**`
            dest_base = dest / posixpath.basename(src)
            create_base = True
        else:
            # `cp src dest_noexists/` copies `src/**` into `dest_noexists/**`
            create_base = True
            dest_base = dest
        src_base = src

        paths = [
            (src_file, dest_base / posixpath.relpath(src_file, src_base))
            for src_file in files.paths
        ]

    if not dest_base.is_dir():
        if make_dirs:
            dest_base.parent.mkdir(exist_ok=True, parents=True)
        elif create_base:
            dest_base.mkdir(exist_ok=True)
        else:
            raise CLIError(Messages.E021, dest_base)
    #     print(
    #         f"""
    #            {paths=}
    # """.strip()
    #     )

    with Progress(paths) as paths:
        for src_file, dest_file in paths:
            # Determine local destination for this file
            if dest_file.exists() and not overwrite:
                # We should only reach this check for recursive copies, so
                # rather than fail halfway through we just warn
                msg.warn(f"{dest_file} exists, skipping...")
                continue
            body = broker_models.Downloading(target=src_file)
            try:
                content = broker_client.files.download(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="download"), e)

            # Save content to local file
            dest_file.parent.mkdir(exist_ok=True, parents=True)
            if verbose:
                msg.info(f"writing {src_file} -> {dest_file}")
            with open(dest_file, "wb") as f:
                f.write(content.read())


def _copy_on_remote(
    *,
    src: str,
    dest: str,
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
) -> ty.Optional[str]:
    if recurse:
        body = broker_models.Copying(
            src=src, dest=dest, make_dirs=make_dirs, overwrite=overwrite
        )
        try:
            plan = broker_client.files.plan_directory_copy(body)
        except BrokerError as e:
            raise CLIError(Messages.E022, e)
        copies = plan.copy_
    else:
        copies = [
            broker_models.Copying(
                src=src, dest=dest, make_dirs=make_dirs, overwrite=overwrite
            )
        ]
    with Progress(copies) as copies:
        for file_copy_plan in copies:
            body = broker_models.Copying(
                src=file_copy_plan.src,
                dest=file_copy_plan.dest,
                make_dirs=make_dirs,
                overwrite=overwrite,
            )
            try:
                broker_client.files.copy(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="copy"), e)



from httpx import (
    ConnectError,
    ConnectTimeout,
    HTTPError,
    HTTPStatusError,
    InvalidURL,
    ReadTimeout,
    RequestError,
)

from . import ty
from .messages import Messages
from .prodigy_teams_broker_sdk.errors import BrokerError
from .prodigy_teams_pam_sdk import errors as ProdigyTeamsErrors

HTTPXErrors = (
    ConnectError,
    ConnectTimeout,
    HTTPError,
    HTTPStatusError,
    InvalidURL,
    RequestError,
    ReadTimeout,
)


class ProdigyTeamsError(Exception):
    def __init__(self, message: str = "") -> None:
        self.message = message
        super().__init__(self.message)


class ProdigyTeamsIDTokenError(ProdigyTeamsError):
    pass


class CLIError(Exception):
    def __init__(self, title: str, text: ty.Optional[ty.Any] = None) -> None:
        self.title = title
        self.text = text
        self.message = self.title + (f"\n{self.text}" if self.text else "")
        super().__init__(self.message)


class ProdigyTeamsParseSecretsError(ProdigyTeamsError):
    def __init__(
        self, secrets_file: ty.Optional[ty.Path], error: ty.Optional[Exception] = None
    ) -> None:
        path_details = (
            f" (`{str(secrets_file.absolute())}`)" if secrets_file is not None else ""
        )
        error_details = f"\n\n{str(error)}" if error is not None else ""
        self.message = Messages.E045.format(
            path_details=path_details, error_details=error_details
        )

    def __str__(self) -> str:
        return self.message


class RecipeBuildMetaFailed(ProdigyTeamsError):
    def __init__(self, package_name: str, stdout: str, stderr: str) -> None:

        self.package_name = package_name
        self.stdout = stdout
        self.stderr = stderr

        self.message = Messages.E048.format(
            package=package_name,
        )

    def __str__(self) -> str:
        return self.message


__all__ = [
    "CLIError",
    "ConnectError",
    "ConnectTimeout",
    "HTTPError",
    "HTTPStatusError",
    "HTTPXErrors",
    "InvalidURL",
    "ProdigyTeamsError",
    "ProdigyTeamsErrors",
    "ProdigyTeamsIDTokenError",
    "ProdigyTeamsParseSecretsError",
    "RequestError",
    "BrokerError",
]



from typing import (
    IO,
    Any,
    Dict,
    Generic,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

from pydantic import BaseModel, conint
from pydantic.generics import GenericModel

T = TypeVar("T")


class Page(GenericModel, Generic[T]):
    items: Sequence[T]
    total: int
    page: conint(ge=1)  # type: ignore
    size: conint(ge=1)  # type: ignore


__all__ = [
    "Any",
    "BaseModel",
    "cast",
    "Dict",
    "IO",
    "List",
    "Literal",
    "NamedTuple",
    "Optional",
    "Type",
    "Union",
    "UUID",
]



from contextlib import ExitStack, contextmanager
from contextvars import ContextVar

from .. import auth, config, ty

ROOT_CONFIG: ContextVar[ty.Optional[config.RootConfig]] = ContextVar(
    "root_config", default=None
)
AUTH_STATE: ContextVar[ty.Optional[auth.AuthState]] = ContextVar(
    "cli_auth", default=None
)
SAVED_SETTINGS: ContextVar[ty.Optional[config.SavedSettings]] = ContextVar(
    "saved_settings", default=None
)
SECRETS: ContextVar[ty.Optional[auth.Secrets]] = ContextVar("secrets", default=None)


def get_root_cfg() -> config.RootConfig:
    """
    Returns the current CLIContext. This can be used in CLI commands access "global" state
    which depends on the environment. During normal CLI operation, the context is provided
    by the __main__ entrypoint. During testing, it can be set manually.
    """
    context = ROOT_CONFIG.get()
    if context is None:
        raise RuntimeError("No CLI context set.")
    return context


def get_auth_state() -> auth.AuthState:
    cli_auth = AUTH_STATE.get()
    if cli_auth is None:
        ctx = get_root_cfg()
        cli_auth = auth.AuthStateImpl(ctx)

        AUTH_STATE.set(cli_auth)

    return cli_auth


def get_saved_settings() -> config.SavedSettings:
    saved_settings = SAVED_SETTINGS.get()
    if saved_settings is None:
        root_cfg = get_root_cfg()
        saved_settings = config.SavedSettings.from_file(root_cfg.saved_settings_path)

        SAVED_SETTINGS.set(saved_settings)

    return saved_settings


def get_secrets() -> auth.Secrets:
    secrets = SECRETS.get()
    if secrets is None:
        root_cfg = get_root_cfg()
        secrets = auth.FileSecrets.load(root_cfg.secrets_path)
        SECRETS.set(secrets)
    return secrets


def clear_state():
    ROOT_CONFIG.set(None)
    AUTH_STATE.set(None)
    SAVED_SETTINGS.set(None)
    SECRETS.set(None)


class UseSystemConfig:
    pass


@contextmanager
def cli_context(
    config_dir: ty.Union[ty.Path, UseSystemConfig] = UseSystemConfig(),
    auth: ty.Optional[auth.AuthState] = None,
    saved_settings: ty.Optional[config.SavedSettings] = None,
    secrets: ty.Optional[auth.Secrets] = None,
    inherit_context: bool = False,
):
    if config_dir is UseSystemConfig or isinstance(config_dir, UseSystemConfig):
        config_dir = config.global_config_dir()

    assignments = [
        (ROOT_CONFIG, config.RootConfig(config_dir=config_dir)),
        (AUTH_STATE, auth),
        (SAVED_SETTINGS, saved_settings),
        (SECRETS, secrets),
    ]
    with ExitStack() as cleanup:
        for var, value in assignments:
            if value is not None or not inherit_context:
                token = var.set(value)
                cleanup.callback(var.reset, token)
        yield



import builtins
import json

from radicli.util import Arg
from wasabi import msg

from .. import ty
from ..cli import cli
from ..errors import CLIError, ProdigyTeamsErrors
from ..messages import Messages
from ..prodigy_teams_pam_sdk.models import AssetCreating, AssetDetail, AssetSummary
from ..query import resolve_asset, resolve_asset_id
from ..ui import print_info_table, print_table_with_select
from ..util import resolve_remote_path
from ._state import get_auth_state


@cli.subcommand(
    "assets",
    "list",
    # fmt: off
    select=Arg("--select", help=Messages.select.format(opts=list(AssetSummary.__fields__))),
    as_json=Arg("--json", help=Messages.as_json),
    # fmt: on
)
def list(
    select: ty.List[str] = ["id", "name", "kind"], as_json: bool = False
) -> ty.Sequence[AssetSummary]:
    """List all assets on the cluster registered with Prodigy Teams"""
    client = get_auth_state().client
    res = client.asset.all()
    print_table_with_select(res.items, select=select, as_json=as_json)
    return res.items


@cli.subcommand(
    "assets",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="asset")),
    project_id=Arg(help=Messages.project_id.format(noun="action")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="action")),
    select=Arg(
        "--select",
        help=Messages.select.format(opts=builtins.list(AssetDetail.__fields__)),
    ),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(
    name_or_id: ty.StrOrUUID,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
    select: ty.Optional[ty.List[str]] = None,
    as_json: bool = False,
) -> AssetDetail:
    """
    Get detailed info for an asset uploaded to the cluster and registered
    with Prodigy Teams
    """
    res = resolve_asset(name_or_id, broker_id=cluster_id, project_id=project_id)
    print_info_table(res, as_json=as_json, select=select)
    return res


@cli.subcommand(
    "assets",
    "create",
    name=Arg(help=Messages.name.format(noun="asset")),
    kind=Arg("--kind", help=Messages.asset_kind),
    path=Arg(help=Messages.path.format(noun="asset")),
    version=Arg("--version", help=Messages.version.format(noun="asset")),
    meta=Arg("--meta", help=Messages.asset_meta),
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok),
)
def create(
    kind: str,
    name: str,
    path: str,
    version: str = "0.0.0",
    meta: str = "{}",
    exists_ok: bool = False,
) -> ty.Optional[ty.UUID]:
    """
    Create an asset on the cluster and register it with Prodigy Teams. Assets
    point to files or directories you control. The Prodigy Teams server only has
    a reference to them. This command doesn't transfer any data. See `ptc files`
    for utilities to transfer files to and from your cluster
    """
    try:
        _ = json.loads(meta)
    except ValueError:
        raise CLIError(Messages.E121, str(meta))
    auth = get_auth_state()
    client = auth.client
    resolved_path = resolve_remote_path(client, path, auth.broker_host)
    try:
        asset = client.asset.create(
            AssetCreating(
                broker_id=auth.broker_id,
                name=name,
                kind=kind,
                version=version,
                meta=json.loads(meta),
                path=resolved_path,
            )
        )
    except ProdigyTeamsErrors.AssetExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="asset", name=name))
            return None
        raise CLIError(Messages.E002.format(noun="asset", name=name))
    assert asset.path is not None
    msg.good(Messages.T002.format(noun="asset", name=asset.name), resolved_path)
    return asset.id


@cli.subcommand(
    "assets",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="asset")),
    project_id=Arg(help=Messages.project_id.format(noun="asset")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="asset")),
)
def delete(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Delete an asset registered with Prodigy Teams"""
    asset_id = resolve_asset_id(name_or_id, broker_id=cluster_id, project_id=project_id)
    auth = get_auth_state()
    try:
        auth.client.asset.delete(id=asset_id)
    except ProdigyTeamsErrors.AssetNotFound:
        raise CLIError(Messages.E006.format(noun="asset", name=name_or_id))
    except ProdigyTeamsErrors.AssetForbiddenDelete:
        raise CLIError(Messages.E007.format(noun="asset", name=name_or_id))
    else:
        msg.good(Messages.T003.format(noun="asset", name=name_or_id))
    return asset_id
